#include "jpegdecoder.h"

int Jpeg::y_width = 0;
int Jpeg::cb_width = 0;
int Jpeg::cr_width = 0;
unsigned char *Jpeg::y_pointer = 0;
unsigned char *Jpeg::cb_pointer = 0;
unsigned char *Jpeg::cr_pointer = 0;

struct error_mgr
{
    struct jpeg_error_mgr mgr;
    jmp_buf setjmp_buffer;
};

typedef struct error_mgr* error_ptr;

METHODDEF(void)
error_exit (j_common_ptr cinfo)
{
  error_ptr myerr = (error_ptr) cinfo->err;
  (*cinfo->err->output_message) (cinfo);
  longjmp(myerr->setjmp_buffer, 1);
}

JpegDecoder::JpegDecoder(ConcurrentQueue<Jpeg *, 1024u> *queue, QObject *parent) :
    QObject(parent), m_queue(queue)
{
}

void JpegDecoder::operate()
{
    struct jpeg_decompress_struct cinfo;
    struct error_mgr jerr;
    cinfo.err = jpeg_std_error(&jerr.mgr);
    jerr.mgr.error_exit = &error_exit;
    if (setjmp(jerr.setjmp_buffer)) {
        jpeg_destroy_decompress(&cinfo);
    }
    jpeg_create_decompress(&cinfo);
    forever {
        auto jpeg = m_queue->dequeue();

        jpeg->output = new JSAMPARRAY[3];

        int paddingHeight = (16 - (jpeg->height % 16));
        int paddingWidth = jpeg->width + (16 - jpeg->width % 16);

        jpeg->padding = new JSAMPLE[3 * paddingHeight * paddingWidth];

        jpeg->rows = new JSAMPROW[3 * (jpeg->height + paddingHeight)];
        for (unsigned char component = 0; component < 3; ++component)
            jpeg->output[component] = &jpeg->rows[component * (jpeg->height + paddingHeight)];

        for (int row = 0; row < jpeg->height; ++row)
        {
            jpeg->output[0][row] = &Jpeg::y_pointer[jpeg->x + (jpeg->y + row) * Jpeg::y_width];
            jpeg->output[1][row] = &Jpeg::cb_pointer[jpeg->x + (jpeg->y + row) * Jpeg::cb_width];
            jpeg->output[2][row] = &Jpeg::cr_pointer[jpeg->x + (jpeg->y + row) * Jpeg::cr_width];
        }
        for (int row = 0; row < paddingHeight; ++row)
        {
            jpeg->output[0][jpeg->height + row] = &jpeg->padding[row * paddingWidth];
            jpeg->output[1][jpeg->height + row] = &jpeg->padding[(paddingHeight + row) * paddingWidth];
            jpeg->output[2][jpeg->height + row] = &jpeg->padding[(2 * paddingHeight + row) * paddingWidth];
        }

        jpeg_mem_src(&cinfo, jpeg->input, jpeg->length);
        jpeg_read_header(&cinfo, TRUE);
        cinfo.dct_method = JDCT_FASTEST;
        cinfo.raw_data_out = TRUE;
        jpeg_start_decompress(&cinfo);

        JSAMPARRAY rowzero[3];

        for(int component = 0; component < cinfo.output_components; component++)
            rowzero[component] = &jpeg->output[component][0];

        while (cinfo.output_scanline < cinfo.output_height) {
            for(int component = 0; component < cinfo.output_components; component++)
                jpeg->output[component] = &rowzero[component][cinfo.output_scanline / (cinfo.max_v_samp_factor * cinfo.comp_info[component].v_samp_factor)];
            jpeg_read_raw_data(&cinfo, jpeg->output, qMax(static_cast<int>(cinfo.output_height - cinfo.output_scanline), cinfo.max_v_samp_factor * DCTSIZE));
        }

        jpeg_finish_decompress(&cinfo);
        emit finishedJpeg();

        delete jpeg;
    }
    jpeg_destroy_decompress(&cinfo);
}
