#ifndef JPEGDECODER_H
#define JPEGDECODER_H

#include <QObject>
#include <QTime>
#include <QDebug>

// jpeglib requires stdlib.h
#include <stdio.h>
#include <jpeglib.h>
// setjmp.h is required for error handling
#include <setjmp.h>

#include "concurrentqueue.h"

struct Jpeg
{
    uint8_t *input;
    std::size_t length;
    JSAMPIMAGE output;
    JSAMPROW *rows;
    JSAMPLE *padding;
    int width;
    int height;
    int x;
    int y;
    static int y_width;
    static int cb_width;
    static int cr_width;
    static unsigned char *y_pointer;
    static unsigned char *cb_pointer;
    static unsigned char *cr_pointer;

    Jpeg(uint8_t *input,
         const std::size_t length,
         const int width,
         const int height,
         const int x,
         const int y) : input(input), length(length), width(width), height(height), x(x), y(y)
    {

    }

    ~Jpeg()
    {
        std::free(input);
        delete[] output;
        delete[] rows;
        delete[] padding;
    }
};

/*!
 * \brief Loads JPEG images
 *
 * Consumes
 */
class JpegDecoder : public QObject
{
    Q_OBJECT
public:
    JpegDecoder(ConcurrentQueue<Jpeg*, 1024u> *queue, QObject *parent = 0);
    ConcurrentQueue<Jpeg*, 1024u> *m_queue;
signals:
    void finishedJpeg();
public slots:
    void operate();
};

#endif // JPEGDECODER_H
