#ifndef VNCRENDERER_H
#define VNCRENDERER_H

#include <memory>
#include <QQuickWindow>
#include <QQuickFramebufferObject>
#include <QOpenGLFramebufferObject>
#include <QOpenGLContext>
#include <QOpenGLFunctions>
#include <QMatrix4x4>
#include <QOpenGLShaderProgram>
#include <QOpenGLBuffer>
#include <QTime>

#include <texture.h>

#include "qvnc.h"

struct Indices
{
    // 4 corner/rectangle and the first and last one duplicated except for the first and last rectangle
    unsigned short i[(VERTEX_BUFFER / 4) * 6 - 2];
    Indices() : i()
    {
        i[0] = 0;
        auto index = 1;
        for(auto value = 1; value < VERTEX_BUFFER; value++) {
            i[index++] = value;
            if(value % 4 == 0 || value % 4 == 3)
                i[index++] = value;
        }
    }
};
const struct Indices ind;

class VncRenderer : public QObject, public QQuickFramebufferObject::Renderer
{
    Q_OBJECT
public:
    VncRenderer(QQuickWindow *window,
                QOpenGLBuffer fillBuffer,
                QOpenGLBuffer copyBuffer,
                QOpenGLBuffer bitmapBuffer,
                QOpenGLBuffer ycbcrBuffer,
                QObject *parent = nullptr);
protected:
    void render() override;
    void synchronize(QQuickFramebufferObject *fbo) override;
    QOpenGLFramebufferObject *createFramebufferObject(const QSize &size) override;
private:
    QQuickWindow *m_window;

    QMatrix4x4 m_fromTexture;
    QMatrix4x4 m_toFramebuffer;
    QMatrix4x4 m_fromFramebuffer;
    QMatrix4x4 m_toTexture;

    QOpenGLBuffer m_fillBuffer;
    QOpenGLBuffer m_copyBuffer;
    QOpenGLBuffer m_bitmapBuffer;
    QOpenGLBuffer m_ycbcrBuffer;

    std::shared_ptr<Texture> m_rgbTexture;
    std::shared_ptr<Texture> m_yTexture;
    std::shared_ptr<Texture> m_cbTexture;
    std::shared_ptr<Texture> m_crTexture;

    unsigned int m_fillCount;
    unsigned int m_copyCount;
    unsigned int m_bitmapCount;
    unsigned int m_ycbcrCount;

    QOpenGLBuffer m_indices;
    QOpenGLShaderProgram m_fillProgram;
    QOpenGLShaderProgram m_bitmapProgram;
    QOpenGLShaderProgram m_ycbcrProgram;
    QOpenGLShaderProgram m_copyProgram;
    GLuint m_textureTarget;
    QOpenGLFunctions *m_gl;

    QOpenGLFramebufferObject *m_framebufferObject;

    bool m_clear;

    int frameCount;
    QTime frameTime;
private slots:
    void handleDebugMessage(const QOpenGLDebugMessage &debugMessage);
signals:
    void rendered();
};

#endif // VNCRENDERER_H
